all_icons = ["address-book", "address-book-o", "address-card", "address-card-o", "bandcamp", "bath", "bathtub",
             "drivers-license", "drivers-license-o", "eercast", "envelope-open", "envelope-open-o", "etsy",
             "free-code-camp", "grav", "handshake-o", "id-badge", "id-card", "id-card-o", "imdb", "linode", "meetup",
             "microchip", "podcast", "quora", "ravelry", "s15", "shower", "snowflake-o", "superpowers", "telegram",
             "thermometer", "thermometer-0", "thermometer-1", "thermometer-2", "thermometer-3", "thermometer-4",
             "thermometer-empty", "thermometer-full", "thermometer-half", "thermometer-quarter",
             "thermometer-three-quarters", "times-rectangle", "times-rectangle-o", "user-circle", "user-circle-o",
             "user-o", "vcard", "vcard-o", "window-close", "window-close-o", "window-maximize", "window-minimize",
             "window-restore", "wpexplorer", "address-book", "address-book-o", "address-card", "address-card-o",
             "adjust", "american-sign-language-interpreting", "anchor", "archive", "area-chart", "arrows", "arrows-h",
             "arrows-v", "asl-interpreting", "assistive-listening-systems", "asterisk", "at", "audio-description",
             "automobile", "balance-scale", "ban", "bank", "bar-chart", "bar-chart-o", "barcode", "bars", "bath",
             "bathtub", "battery", "battery-0", "battery-1", "battery-2", "battery-3", "battery-4", "battery-empty",
             "battery-full", "battery-half", "battery-quarter", "battery-three-quarters", "bed", "beer", "bell",
             "bell-o", "bell-slash", "bell-slash-o", "bicycle", "binoculars", "birthday-cake", "blind", "bluetooth",
             "bluetooth-b", "bolt", "bomb", "book", "bookmark", "bookmark-o", "braille", "briefcase", "bug", "building",
             "building-o", "bullhorn", "bullseye", "bus", "cab", "calculator", "calendar", "calendar-check-o",
             "calendar-minus-o", "calendar-o", "calendar-plus-o", "calendar-times-o", "camera", "camera-retro", "car",
             "caret-square-o-down", "caret-square-o-left", "caret-square-o-right", "caret-square-o-up",
             "cart-arrow-down", "cart-plus", "cc", "certificate", "check", "check-circle", "check-circle-o",
             "check-square", "check-square-o", "child", "circle", "circle-o", "circle-o-notch", "circle-thin",
             "clock-o", "clone", "close", "cloud", "cloud-download", "cloud-upload", "code", "code-fork", "coffee",
             "cog", "cogs", "comment", "comment-o", "commenting", "commenting-o", "comments", "comments-o", "compass",
             "copyright", "creative-commons", "credit-card", "credit-card-alt", "crop", "crosshairs", "cube", "cubes",
             "cutlery", "dashboard", "database", "deaf", "deafness", "desktop", "diamond", "dot-circle-o", "download",
             "drivers-license", "drivers-license-o", "edit", "ellipsis-h", "ellipsis-v", "envelope", "envelope-o",
             "envelope-open", "envelope-open-o", "envelope-square", "eraser", "exchange", "exclamation",
             "exclamation-circle", "exclamation-triangle", "external-link", "external-link-square", "eye", "eye-slash",
             "eyedropper", "fax", "feed", "female", "fighter-jet", "file-archive-o", "file-audio-o", "file-code-o",
             "file-excel-o", "file-image-o", "file-movie-o", "file-pdf-o", "file-photo-o", "file-picture-o",
             "file-powerpoint-o", "file-sound-o", "file-video-o", "file-word-o", "file-zip-o", "film", "filter", "fire",
             "fire-extinguisher", "flag", "flag-checkered", "flag-o", "flash", "flask", "folder", "folder-o",
             "folder-open", "folder-open-o", "frown-o", "futbol-o", "gamepad", "gavel", "gear", "gears", "gift",
             "glass", "globe", "graduation-cap", "group", "hand-grab-o", "hand-lizard-o", "hand-paper-o",
             "hand-peace-o", "hand-pointer-o", "hand-rock-o", "hand-scissors-o", "hand-spock-o", "hand-stop-o",
             "handshake-o", "hard-of-hearing", "hashtag", "hdd-o", "headphones", "heart", "heart-o", "heartbeat",
             "history", "home", "hotel", "hourglass", "hourglass-1", "hourglass-2", "hourglass-3", "hourglass-end",
             "hourglass-half", "hourglass-o", "hourglass-start", "i-cursor", "id-badge", "id-card", "id-card-o",
             "image", "inbox", "industry", "info", "info-circle", "institution", "key", "keyboard-o", "language",
             "laptop", "leaf", "legal", "lemon-o", "level-down", "level-up", "life-bouy", "life-buoy", "life-ring",
             "life-saver", "lightbulb-o", "line-chart", "location-arrow", "lock", "low-vision", "magic", "magnet",
             "mail-forward", "mail-reply", "mail-reply-all", "male", "map", "map-marker", "map-o", "map-pin",
             "map-signs", "meh-o", "microchip", "microphone", "microphone-slash", "minus", "minus-circle",
             "minus-square", "minus-square-o", "mobile", "mobile-phone", "money", "moon-o", "mortar-board",
             "motorcycle", "mouse-pointer", "music", "navicon", "newspaper-o", "object-group", "object-ungroup",
             "paint-brush", "paper-plane", "paper-plane-o", "paw", "pencil", "pencil-square", "pencil-square-o",
             "percent", "phone", "phone-square", "photo", "picture-o", "pie-chart", "plane", "plug", "plus",
             "plus-circle", "plus-square", "plus-square-o", "podcast", "power-off", "print", "puzzle-piece", "qrcode",
             "question", "question-circle", "question-circle-o", "quote-left", "quote-right", "random", "recycle",
             "refresh", "registered", "remove", "reorder", "reply", "reply-all", "retweet", "road", "rocket", "rss",
             "rss-square", "s15", "search", "search-minus", "search-plus", "send", "send-o", "server", "share",
             "share-alt", "share-alt-square", "share-square", "share-square-o", "shield", "ship", "shopping-bag",
             "shopping-basket", "shopping-cart", "shower", "sign-in", "sign-language", "sign-out", "signal", "signing",
             "sitemap", "sliders", "smile-o", "snowflake-o", "soccer-ball-o", "sort", "sort-alpha-asc",
             "sort-alpha-desc", "sort-amount-asc", "sort-amount-desc", "sort-asc", "sort-desc", "sort-down",
             "sort-numeric-asc", "sort-numeric-desc", "sort-up", "space-shuttle", "spinner", "spoon", "square",
             "square-o", "star", "star-half", "star-half-empty", "star-half-full", "star-half-o", "star-o",
             "sticky-note", "sticky-note-o", "street-view", "suitcase", "sun-o", "support", "tablet", "tachometer",
             "tag", "tags", "tasks", "taxi", "television", "terminal", "thermometer", "thermometer-0", "thermometer-1",
             "thermometer-2", "thermometer-3", "thermometer-4", "thermometer-empty", "thermometer-full",
             "thermometer-half", "thermometer-quarter", "thermometer-three-quarters", "thumb-tack", "thumbs-down",
             "thumbs-o-down", "thumbs-o-up", "thumbs-up", "ticket", "times", "times-circle", "times-circle-o",
             "times-rectangle", "times-rectangle-o", "tint", "toggle-down", "toggle-left", "toggle-off", "toggle-on",
             "toggle-right", "toggle-up", "trademark", "trash", "trash-o", "tree", "trophy", "truck", "tty", "tv",
             "umbrella", "universal-access", "university", "unlock", "unlock-alt", "unsorted", "upload", "user",
             "user-circle", "user-circle-o", "user-o", "user-plus", "user-secret", "user-times", "users", "vcard",
             "vcard-o", "video-camera", "volume-control-phone", "volume-down", "volume-off", "volume-up", "warning",
             "wheelchair", "wheelchair-alt", "wifi", "window-close", "window-close-o", "window-maximize",
             "window-minimize", "window-restore", "wrench", "american-sign-language-interpreting", "asl-interpreting",
             "assistive-listening-systems", "audio-description", "blind", "braille", "cc", "deaf", "deafness",
             "hard-of-hearing", "low-vision", "question-circle-o", "sign-language", "signing", "tty",
             "universal-access", "volume-control-phone", "wheelchair", "wheelchair-alt", "hand-grab-o", "hand-lizard-o",
             "hand-o-down", "hand-o-left", "hand-o-right", "hand-o-up", "hand-paper-o", "hand-peace-o",
             "hand-pointer-o", "hand-rock-o", "hand-scissors-o", "hand-spock-o", "hand-stop-o", "thumbs-down",
             "thumbs-o-down", "thumbs-o-up", "thumbs-up", "ambulance", "automobile", "bicycle", "bus", "cab", "car",
             "fighter-jet", "motorcycle", "plane", "rocket", "ship", "space-shuttle", "subway", "taxi", "train",
             "truck", "wheelchair", "wheelchair-alt", "genderless", "intersex", "mars", "mars-double", "mars-stroke",
             "mars-stroke-h", "mars-stroke-v", "mercury", "neuter", "transgender", "transgender-alt", "venus",
             "venus-double", "venus-mars", "file", "file-archive-o", "file-audio-o", "file-code-o", "file-excel-o",
             "file-image-o", "file-movie-o", "file-o", "file-pdf-o", "file-photo-o", "file-picture-o",
             "file-powerpoint-o", "file-sound-o", "file-text", "file-text-o", "file-video-o", "file-word-o",
             "file-zip-o", "circle-o-notch", "cog", "gear", "refresh", "spinner", "check-square", "check-square-o",
             "circle", "circle-o", "dot-circle-o", "minus-square", "minus-square-o", "plus-square", "plus-square-o",
             "square", "square-o", "cc-amex", "cc-diners-club", "cc-discover", "cc-jcb", "cc-mastercard", "cc-paypal",
             "cc-stripe", "cc-visa", "credit-card", "credit-card-alt", "google-wallet", "paypal", "area-chart",
             "bar-chart", "bar-chart-o", "line-chart", "pie-chart", "bitcoin", "btc", "cny", "dollar", "eur", "euro",
             "gbp", "gg", "gg-circle", "ils", "inr", "jpy", "krw", "money", "rmb", "rouble", "rub", "ruble", "rupee",
             "shekel", "sheqel", "try", "turkish-lira", "usd", "viacoin", "won", "yen", "align-center", "align-justify",
             "align-left", "align-right", "bold", "chain", "chain-broken", "clipboard", "columns", "copy", "cut",
             "dedent", "eraser", "file", "file-o", "file-text", "file-text-o", "files-o", "floppy-o", "font", "header",
             "indent", "italic", "link", "list", "list-alt", "list-ol", "list-ul", "outdent", "paperclip", "paragraph",
             "paste", "repeat", "rotate-left", "rotate-right", "save", "scissors", "strikethrough", "subscript",
             "superscript", "table", "text-height", "text-width", "th", "th-large", "th-list", "underline", "undo",
             "unlink", "angle-double-down", "angle-double-left", "angle-double-right", "angle-double-up", "angle-down",
             "angle-left", "angle-right", "angle-up", "arrow-circle-down", "arrow-circle-left", "arrow-circle-o-down",
             "arrow-circle-o-left", "arrow-circle-o-right", "arrow-circle-o-up", "arrow-circle-right",
             "arrow-circle-up", "arrow-down", "arrow-left", "arrow-right", "arrow-up", "arrows", "arrows-alt",
             "arrows-h", "arrows-v", "caret-down", "caret-left", "caret-right", "caret-square-o-down",
             "caret-square-o-left", "caret-square-o-right", "caret-square-o-up", "caret-up", "chevron-circle-down",
             "chevron-circle-left", "chevron-circle-right", "chevron-circle-up", "chevron-down", "chevron-left",
             "chevron-right", "chevron-up", "exchange", "hand-o-down", "hand-o-left", "hand-o-right", "hand-o-up",
             "long-arrow-down", "long-arrow-left", "long-arrow-right", "long-arrow-up", "toggle-down", "toggle-left",
             "toggle-right", "toggle-up", "arrows-alt", "backward", "compress", "eject", "expand", "fast-backward",
             "fast-forward", "forward", "pause", "pause-circle", "pause-circle-o", "play", "play-circle",
             "play-circle-o", "random", "step-backward", "step-forward", "stop", "stop-circle", "stop-circle-o",
             "youtube-play", "500px", "adn", "amazon", "android", "angellist", "apple", "bandcamp", "behance",
             "behance-square", "bitbucket", "bitbucket-square", "bitcoin", "black-tie", "bluetooth", "bluetooth-b",
             "btc", "buysellads", "cc-amex", "cc-diners-club", "cc-discover", "cc-jcb", "cc-mastercard", "cc-paypal",
             "cc-stripe", "cc-visa", "chrome", "codepen", "codiepie", "connectdevelop", "contao", "css3", "dashcube",
             "delicious", "deviantart", "digg", "dribbble", "dropbox", "drupal", "edge", "eercast", "empire", "envira",
             "etsy", "expeditedssl", "fa", "facebook", "facebook-f", "facebook-official", "facebook-square", "firefox",
             "first-order", "flickr", "font-awesome", "fonticons", "fort-awesome", "forumbee", "foursquare",
             "free-code-camp", "ge", "get-pocket", "gg", "gg-circle", "git", "git-square", "github", "github-alt",
             "github-square", "gitlab", "gittip", "glide", "glide-g", "google", "google-plus", "google-plus-circle",
             "google-plus-official", "google-plus-square", "google-wallet", "gratipay", "grav", "hacker-news", "houzz",
             "html5", "imdb", "instagram", "internet-explorer", "ioxhost", "joomla", "jsfiddle", "lastfm",
             "lastfm-square", "leanpub", "linkedin", "linkedin-square", "linode", "linux", "maxcdn", "meanpath",
             "medium", "meetup", "mixcloud", "modx", "odnoklassniki", "odnoklassniki-square", "opencart", "openid",
             "opera", "optin-monster", "pagelines", "paypal", "pied-piper", "pied-piper-alt", "pied-piper-pp",
             "pinterest", "pinterest-p", "pinterest-square", "product-hunt", "qq", "quora", "ra", "ravelry", "rebel",
             "reddit", "reddit-alien", "reddit-square", "renren", "resistance", "safari", "scribd", "sellsy",
             "share-alt", "share-alt-square", "shirtsinbulk", "simplybuilt", "skyatlas", "skype", "slack", "slideshare",
             "snapchat", "snapchat-ghost", "snapchat-square", "soundcloud", "spotify", "stack-exchange",
             "stack-overflow", "steam", "steam-square", "stumbleupon", "stumbleupon-circle", "superpowers", "telegram",
             "tencent-weibo", "themeisle", "trello", "tripadvisor", "tumblr", "tumblr-square", "twitch", "twitter",
             "twitter-square", "usb", "viacoin", "viadeo", "viadeo-square", "vimeo", "vimeo-square", "vine", "vk",
             "wechat", "weibo", "weixin", "whatsapp", "wikipedia-w", "windows", "wordpress", "wpbeginner", "wpexplorer",
             "wpforms", "xing", "xing-square", "y-combinator", "y-combinator-square", "yahoo", "yc", "yc-square",
             "yelp", "yoast", "youtube", "youtube-play", "youtube-square"]



import os
import uuid

from chemotion_api.connection import Connection
from chemotion_api.elements.attachments import Attachment
import magic

class UniqueDict(dict):

    def __setitem__(self, key, value):
        # optional processing here
        org_key = key
        idx = 1
        while key in self:
            idx += 1
            key = f'{org_key}_{idx}'
        super().__setitem__(key, value)

class SenderBox:
    def __init__(self, con: Connection, json_data: dict):
        self._content = None
        self._con = con
        self._json_data = json_data

    @property
    def attachments(self) -> UniqueDict[str:UniqueDict[str:Attachment]] :
        if self._content is None:
            self.load()
        return self._content


    def load(self):
        self._content = UniqueDict()
        page = 0
        content_length = -1
        while content_length != 0:
            page += 1
            res = self._con.get(f'/api/v1/inbox/containers/{self._json_data["id"]}?dataset_page={page}&sort_column=name')
            if res.status_code != 200:
                raise ConnectionError(f"Inbox could not be fetched! {res} -> {res.status_code}")
            json_res = res.json()['inbox']
            content_length = len(json_res['children'])
            for child in json_res['children']:
                att_dict = UniqueDict()
                for att in child['attachments']:
                    att_dict[att['filename']] = Attachment(self._con, att)

                self._content[child['name']] = att_dict


class Inbox:
    def __init__(self, con: Connection):
        self._content = None
        self._unsorted_content = None
        self._con = con

    @property
    def sorted_boxs(self) -> UniqueDict[str:SenderBox] :
        if self._content is None:
            self.load()
        return self._content

    @property
    def unsorted_attachments(self) -> UniqueDict[str:Attachment] :
        if self._content is None:
            self.load()
        return self._unsorted_content

    def load(self):
        page = 0
        max_page = 1
        per_page = 20
        json_res = {}
        self._content = UniqueDict()
        self._unsorted_content = UniqueDict()
        while page < max_page:
            page += 1
            res = self._con.get(f'/api/v1/inbox?cnt_only=false&page={page}&per_page={per_page}&sort_column=name')
            if res.status_code != 200:
                raise ConnectionError(f"Inbox could not be fetched! {res} -> {res.status_code}")
            json_res = res.json()['inbox']
            max_page = 1 + json_res['count'] // per_page
            for child in json_res['children']:
                self._content[child['name']] = SenderBox(self._con, child)
        for child in json_res['unlinked_attachments']:
            self._unsorted_content[child['filename']] = Attachment(self._con, child)

    def upload(self, file_path: str):
        param_name = uuid.uuid4().__str__()
        file_magic = magic.Magic(mime=True)

        files = [(param_name, (os.path.basename(file_path), open(file_path, 'rb'), file_magic.from_file(file_path)))]
        res = self._con.post('/api/v1/attachments/upload_to_inbox', data={}, files=files)
        return res




from chemotion_api.labImotion.items.options import ConditionOperator, LayerColor
from chemotion_api.labImotion.items.validation.registry import SchemaRegistry

generic_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "chemotion://generic/layer/draft-01",
    "title": "Schema for generic layer",
    "type": "object",
    "properties": {
        "wf": {
            "type": "boolean"
        },
        "key": {
            "type": "string",
            "pattern": "^[a-z][a-zA-Z_]{1,20}[a-z]$"
        },
        "cols": {
            "type": "number"
        },
        "color": {
            "type": "string",
            "enum": LayerColor.list()
        },
        "label": {
            "type": "string"
        },
        "style": {
            "type": "string",
            "pattern": "^panel_generic_heading$|^panel_generic_heading_([bui])(?!.*\1)([bui])?(?!.*\1|.*\2)([bui])?$"
        },
        "fields": {
            "type": "array",
            "items": {
                "$ref": "chemotion://generic/field/draft-01"
            }
        },
        "position": {
            "type": "number",
            "minimum": 1
        },
        "timeRecord": {
            "type": "string"
        },
        "cond_fields": {
            "type": "array",
            "items": {
                "$ref": "chemotion://generic/cond_field/draft-01"
            }
        },
        "wf_position": {
            "type": "number",
            "minimum": 0
        },
        "cond_operator": {
            "type": "number",
            "enum": ConditionOperator.list()
        }
    },
    "additionalProperties": False,
    "required": [
        "wf",
        "key",
        "cols",
        "color",
        "label",
        "style",
        "fields",
        "position",
        "timeRecord",
        "wf_position"
    ]
}

SchemaRegistry.instance().register(generic_schema)


import os

import chemotion_api.labImotion.items.validation.schemas.imports
from chemotion_api.labImotion.items.validation.registry import SchemaRegistry
from chemotion_api.labImotion.items.validation.schemas import ALL_SCHEMAS


if __name__ == '__main__':
    A = []
    for module in os.listdir(os.path.join(os.path.dirname(__file__), 'schemas')):
        if module.startswith('schema_') and module[-3:] == '.py':
            A.append(f'import chemotion_api.labImotion.items.validation.schemas.{module[:-3]}')

    with open(os.path.join(os.path.dirname(__file__), 'schemas/imports.py'), 'w') as f:
        f.write('\n'.join(A))
    exit()


def validate_selection_options(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/select_option/draft-01')
    validator.validate(json_to_test)


def validate_generic_element(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/element/draft-01')
    validator.validate(json_to_test)


def validate_generic_dataset(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/dataset/draft-01')
    validator.validate(json_to_test)


def validate_generic_segment(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/segment/draft-01')
    validator.validate(json_to_test)


def validate_generic_properties(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/properties/draft-01')
    validator.validate(json_to_test)


def validate_generic_dataset_properties(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/dataset_properties/draft-01')
    validator.validate(json_to_test)


def validate_generic_segment_properties(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/segment_properties/draft-01')
    validator.validate(json_to_test)


def validate_generic_layer(json_to_test: dict):
    validator = SchemaRegistry.instance().validator_for('chemotion://generic/layer/draft-01')
    validator.validate(json_to_test)



from dataclasses import fields
from typing import Literal, Optional, Self

from chemotion_api.search.search_helper import SearchHelperBase
from chemotion_api.search.utils import EnumMatch, EnumConnector, new_base_field


def field_details(field_name: Literal[
    'name', 'short_label', 'temperature', 'status', 'conditions', 'rxno', 'duration', 'content']):
    unit = ''
    if field_name == 'name':
        filed_type = {
            "column": "name",
            "label": "Name",
            "type": "text",
            "advanced": True
        }
    elif field_name == 'short_label':
        filed_type = {
            "column": "short_label",
            "label": "Short Label",
            "type": "text",
            "advanced": True
        }

    elif field_name == 'status':
        filed_type = {
            "column": "status",
            "label": "Status",
            "type": "select",
            "option_layers": "statusOptions",
            "advanced": True
        }

    elif field_name == 'conditions':
        filed_type = {
            "column": "conditions",
            "label": "Conditions",
            "type": "text",
            "advanced": True
        }

    elif field_name == 'temperature':
        filed_type = {
            "column": "temperature",
            "label": "Temperature",
            "type": "system-defined",
            "option_layers": "temperature",
            "info": "Only numbers are allowed",
            "advanced": True
        }
        unit = "°C"

    elif field_name == 'duration':
        filed_type = {
            "column": "duration",
            "label": "Duration",
            "type": "system-defined",
            "option_layers": "duration",
            "info": "Only numbers are allowed",
            "advanced": True
        }
        unit = "Second(s)"

    elif field_name == 'rxno':
        filed_type = {
            "column": "rxno",
            "label": "Type",
            "type": "rxnos",
            "advanced": True
        }

    elif field_name == 'content':
        filed_type = {
            "column": "content",
            "label": "Private Note",
            "type": "text",
            "advanced": True
        }

    else:
        raise ValueError(f'Field "{field_name}" not supported.')

    return {
        'field': filed_type,
        'unit': unit
    }


class ReactionSearcher(SearchHelperBase):
    element_type_res = "reactions"
    element_type = "reaction"
    def add_search_condition(self, match: EnumMatch, connector: EnumConnector, name: Optional[str] = None,
                             short_label: Optional[str] = None, status: Optional[str] = None,
                             conditions: Optional[str] = None, reaction_type: Optional[str] = None,
                             temperature_in_c: Optional[float] = None, duration_in_sec: Optional[int] = None,
                             private_note: Optional[str] = None):
        field_values = {"name": name, "short_label": short_label, "temperature": temperature_in_c,
                        "status": status, "conditions": conditions, "rxno": reaction_type, "duration": duration_in_sec,
                        "content": private_note}
        return super().add_search_condition(match, connector, **field_values)

    def new_field(self, match: EnumMatch, connector: str, field_name: Literal[
        'name', 'short_label', 'temperature', 'status', 'conditions', 'rxno', 'duration', 'content'],
                  value: str) -> dict:
        return new_base_field(match, connector, value) | {
            "table": "reactions",
            "element_id": 0
        } | field_details(field_name)



from abc import ABC, abstractmethod


class WorkflowInterface(ABC):
    @abstractmethod
    def wf_tree(self):
        raise NotImplementedError()

    @abstractmethod
    def wf_next_options(self, flat=True) -> list:
        raise NotImplementedError()

    @abstractmethod
    def wf_next(self, layer_id: None | str = None, wf_idx: int = 0, wf_id: str = None, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def wf_add_layer(self, layer_label: str, pos: int | None = None):
        raise NotImplementedError()

    @abstractmethod
    def wf_current_workflow(self):
        raise NotImplementedError()



from chemotion_api.labImotion.items.fields.base import BaseField
from chemotion_api.labImotion.items.fields.fields import TextField, CheckboxField, DatetimeField, DatetimeRangeField, \
    DragSampleField, DragMoleculeField, DragElementField, FormulaFieldField, InputGroupField, IntegerField, SelectField, \
    SystemDefinedField, TableField, TextFormulaField, TextareaField, UploadField, DummyField
from chemotion_api.labImotion.items.options import FieldType
from chemotion_api.labImotion.structure import StructureRegistry


def field_factory(sr: StructureRegistry, layer_key: str, json_data) -> BaseField:

    ft = FieldType(json_data.get('type'))
    return _field_factory(sr, layer_key, ft, json_data)


def create_field(sr: StructureRegistry, layer_key: str, field_type: FieldType, label: str, position: int, field: str) -> BaseField:
    return _field_factory(sr, layer_key, field_type, {'type': field_type.value, 'label': label, 'field': field, 'position': position})


def _field_factory(sr: StructureRegistry, layer_key: str, field_type: FieldType, json_data) -> BaseField:
    match field_type:
        case FieldType.DUMMY:
            return DummyField(sr, layer_key, json_data)
        case FieldType.CHECKBOX:
            return CheckboxField(sr, layer_key, json_data)
        case FieldType.DATETIME:
            return DatetimeField(sr, layer_key, json_data)
        case FieldType.DATETIME_RANGE:
            return DatetimeRangeField(sr, layer_key, json_data)
        case FieldType.DRAG_ELEMENT:
            return DragElementField(sr, layer_key, json_data)
        case FieldType.DRAG_MOLECULE:
            return DragMoleculeField(sr, layer_key, json_data)
        case FieldType.DRAG_SAMPLE:
            return DragSampleField(sr, layer_key, json_data)
        case FieldType.FORMULA_FIELD:
            return FormulaFieldField(sr, layer_key, json_data)
        case FieldType.INPUT_GROUP:
            return InputGroupField(sr, layer_key, json_data)
        case FieldType.INTEGER:
            return IntegerField(sr, layer_key, json_data)
        case FieldType.SELECT:
            return SelectField(sr, layer_key, json_data)
        case FieldType.SYSTEM_DEFINED:
            return SystemDefinedField(sr, layer_key, json_data)
        case FieldType.TABLE:
            return TableField(sr, layer_key, json_data)
        case FieldType.TEXT_FORMULA:
            return TextFormulaField(sr, layer_key, json_data)
        case FieldType.TEXT:
            return TextField(sr, layer_key, json_data)
        case FieldType.TEXTAREA:
            return TextareaField(sr, layer_key, json_data)
        case FieldType.UPLOAD:
            return UploadField(sr, layer_key, json_data)

    raise ValueError(f"Field type {FieldType} is not supported")



import os

from jinja2 import Environment, PackageLoader, select_autoescape


def walk_productive_template(url, eln_port, postgres_port, adminer_port, msconvert, spectra, eln, ketchersvc, converter, **kwargs):
    main_root_dir = os.path.join(os.path.dirname(__file__), 'templates/productive')
    for root, dir, files in os.walk(main_root_dir):

        env = Environment(
            loader=PackageLoader("chemotion_api", package_path=root),

            autoescape=select_autoescape()
        )

        context = dict(url=url, eln_port=eln_port, postgres_port=postgres_port, adminer_port=adminer_port,
                       msconvert=msconvert, spectra=spectra, eln=eln, ketchersvc=ketchersvc, converter=converter, **kwargs)

        for f in files:
            rel_root = os.path.relpath(os.path.join(root, f), main_root_dir)
            t = env.get_template(f)
            yield (True, rel_root, t.render(**context))
        for d in dir:
            rel_root = os.path.relpath(os.path.join(root, d), main_root_dir)
            yield (False, rel_root, None)



from chemotion_api.labImotion.items.validation.registry import SchemaRegistry

generic_schema = {
    "$schema": "https://json-schema.org/draft/2020-12/schema",
    "$id": "chemotion://generic/flow_obj/draft-01",
    "title": "Schema for generic element properties",
    "type": "object",
    "properties": {
        "edges": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {
                        "type": "string"
                    },
                    "label": {
                        "type": "string"
                    },
                    "source": {
                        "type": "string"
                    },
                    "target": {
                        "type": "string"
                    },
                    "animated": {
                        "type": "boolean"
                    },
                    "markerEnd": {
                        "type": "object",
                        "properties": {
                            "type": {
                                "type": "string"
                            }
                        },
                        "additionalProperties": False,
                        "required": [
                            "type"
                        ]
                    },
                    "sourceHandle": {},
                    "targetHandle": {}
                },
                "additionalProperties": False,
                "required": [
                    "id",
                    "label",
                    "source",
                    "target",
                    "animated",
                    "markerEnd",
                    "sourceHandle",
                    "targetHandle"
                ]
            }
        },
        "nodes": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "id": {
                        "type": "string"
                    },
                    "data": {
                        "type": "object",
                        "properties": {
                            "label": {
                                "type": "string"
                            },
                            "lKey": {
                                "type": "string"
                            },
                            "layer": {
                                "type": "object",
                                "properties": {
                                    "wf": {
                                        "type": "boolean"
                                    },
                                    "key": {
                                        "type": "string"
                                    },
                                    "cols": {
                                        "type": "number"
                                    },
                                    "color": {
                                        "type": "string"
                                    },
                                    "label": {
                                        "type": "string"
                                    },
                                    "style": {
                                        "type": "string"
                                    },
                                    "fields": {
                                        "type": "array",
                                        "items": {}
                                    },
                                    "position": {
                                        "type": "number"
                                    },
                                    "timeRecord": {
                                        "type": "string"
                                    },
                                    "wf_position": {
                                        "type": "number"
                                    }
                                },
                                "additionalProperties": False,
                                "required": [
                                    "wf",
                                    "key",
                                    "cols",
                                    "color",
                                    "label",
                                    "style",
                                    "fields",
                                    "position",
                                    "timeRecord",
                                    "wf_position"
                                ]
                            }
                        },
                        "required": []
                    },
                    "type": {
                        "type": "string"
                    },
                    "width": {
                        "type": "number"
                    },
                    "height": {
                        "type": "number"
                    },
                    "position": {
                        "type": "object",
                        "properties": {
                            "x": {
                                "type": "number"
                            },
                            "y": {
                                "type": "number"
                            }
                        },
                        "additionalProperties": False,
                        "required": [
                            "x",
                            "y"
                        ]
                    },
                    "deletable": {
                        "type": "boolean"
                    },
                    "positionAbsolute": {
                        "type": "object",
                        "properties": {
                            "x": {
                                "type": "number"
                            },
                            "y": {
                                "type": "number"
                            }
                        },
                        "additionalProperties": False,
                        "required": [
                            "x",
                            "y"
                        ]
                    }
                },
                "additionalProperties": False,
                "required": [
                    "id",
                    "data",
                    "type",
                    "width",
                    "height",
                    "position",
                    "positionAbsolute"
                ]
            }
        },
        "viewport": {
            "type": "object",
            "properties": {
                "x": {
                    "type": "number"
                },
                "y": {
                    "type": "number"
                },
                "zoom": {
                    "type": "number"
                }
            },
            "required": [
                "x",
                "y",
                "zoom"
            ]
        }
    },
    "required": [
        "edges",
        "nodes",
        "viewport"
    ]

}

SchemaRegistry.instance().register(generic_schema)



import re

from chemotion_api.labImotion.items.fields import BaseField
from chemotion_api.labImotion.items.fields.abstract_fields import SystemDefined, Select
from chemotion_api.labImotion.items.fields.sub_fields import SubBaseField, NumberField, TextField as SubTextField, \
    SystemDefinedField as SubSystemDefinedField, LabelField
from chemotion_api.labImotion.items.fields.table_sub_fields import TableSubBaseField, TSDragMolecule, TSDragSample, \
    TSSelect, TSSystemDefined, TSText
from chemotion_api.labImotion.items.options import FieldType, FieldUnit, get_unit, SubFieldType, SubTableFieldType
from chemotion_api.labImotion.structure import StructureRegistry


class DummyField(BaseField):
    field_type = FieldType.DUMMY

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()



class CheckboxField(BaseField):
    field_type = FieldType.CHECKBOX

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class DatetimeField(BaseField):
    field_type = FieldType.DATETIME

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._readonly = False
        self._required = False
        self._parse_json()

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"required {type(value)} not supported")
        self._required = value

    @property
    def readonly(self) -> bool:
        return self._readonly

    @readonly.setter
    def readonly(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"readonly {type(value)} not supported")
        self._readonly = value


class DatetimeRangeField(BaseField):
    field_type = FieldType.DATETIME_RANGE

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class DragElementField(BaseField):
    field_type = FieldType.DRAG_ELEMENT

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class DragMoleculeField(BaseField):
    field_type = FieldType.DRAG_MOLECULE

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class DragSampleField(BaseField):
    field_type = FieldType.DRAG_SAMPLE

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class FormulaFieldField(BaseField):
    field_type = FieldType.FORMULA_FIELD
    required_fields = ["formula", "decimal", "canAdjust"]

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._formula = ""
        self._decimal = 2
        self._can_adjust = True
        self._parse_json()

    @property
    def canAdjust(self) -> bool:
        return self._can_adjust

    @canAdjust.setter
    def canAdjust(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"canAdjust {type(value)} not supported")
        self._can_adjust = value

    @property
    def decimal(self) -> str:
        return str(self._decimal)

    @decimal.setter
    def decimal(self, value: str | int):
        if isinstance(value, int):
            value = str(value)
        if re.match(r"^[0-9]+$", value) is None:
            raise TypeError(f"decimal {type(value)} not supported. Must be a positive integer as string.")
        self._decimal = value

    @property
    def formula(self) -> str:
        return self._formula

    @formula.setter
    def formula(self, value: str):
        if value == '':
            self._formula = '0'
            return
        structure = self._sr.get_structure()[self._layer_key]
        if not isinstance(value, str):
            raise TypeError(f"formula {type(value)} not supported")
        for variable in re.split(r'[+\-*/]', value):
            variable = variable.strip('() ')
            try:
                float(variable)
            except ValueError:
                if structure.get(variable) not in [FieldType.INTEGER, FieldType.SYSTEM_DEFINED]:
                    raise ValueError(f"${variable} not allowed! Only numbers (integer or floating '12.3'), field names of {FieldType.INTEGER} or {FieldType.SYSTEM_DEFINED} in the same layer or operations like addition (+), subtraction (-), multiplication (*), and division (/)")

        self._formula = value


class InputGroupField(BaseField):
    field_type = FieldType.INPUT_GROUP

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._sub_fields = []
        self._parse_json()

    @property
    def sub_fields(self) -> list:
        return self._sub_fields

    @sub_fields.setter
    def sub_fields(self, value: list):
        new_val = []
        for entry in value:
            if isinstance(entry, SubBaseField):
                new_val.append(entry)
            elif isinstance(entry, dict) and entry.get('type') in SubFieldType.list():
                sf_type = SubFieldType(entry['type'])
                new_val.append(self._factory_subfield(sf_type, entry))
            else:
                raise TypeError("Value entries must be of type SubBaseField or dict")

        self._sub_fields = new_val



    def add_subfield(self, sf_type: SubFieldType) -> NumberField | SubTextField | SubSystemDefinedField | LabelField:
        self.sub_fields.append(self._factory_subfield(sf_type))
        return self.sub_fields[-1]

    def _factory_subfield(self, sf_type: SubFieldType,
                          data=None) -> NumberField | SubTextField | SubSystemDefinedField | LabelField:
        if data is None:
            data = {}
        if sf_type == SubFieldType.TEXT:
            return SubTextField(self._sr, self._layer_key, data)
        elif sf_type == SubFieldType.LABEL:
            return LabelField(self._sr, self._layer_key, data)
        elif sf_type == SubFieldType.NUMBER:
            return NumberField(self._sr, self._layer_key, data)
        elif sf_type == SubFieldType.SYSTEM_DEFINED:
            return SubSystemDefinedField(self._sr, self._layer_key, data)


class IntegerField(BaseField):
    field_type = FieldType.INTEGER

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._required = False
        self._parse_json()

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"required {type(value)} not supported")
        self._required = value


class SelectField(BaseField, Select):
    field_type = FieldType.SELECT
    required_fields = ['option_layers']

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        Select.__init__(self, sr, json_data)
        self._parse_json()


class SystemDefinedField(BaseField, SystemDefined):
    field_type = FieldType.SYSTEM_DEFINED
    required_fields = ['option_layers', 'value_system']

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        SystemDefined.__init__(self, json_data)
        self._parse_json()


class TableField(BaseField):
    field_type = FieldType.TABLE

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._sub_fields = []
        self._parse_json()

    @property
    def sub_fields(self) -> list:
        return self._sub_fields

    @sub_fields.setter
    def sub_fields(self, value: list):
        new_val = []
        for entry in value:
            if isinstance(entry, TableSubBaseField):
                new_val.append(entry)
            elif isinstance(entry, dict) and entry.get('type') in SubTableFieldType.list():
                sf_type = SubTableFieldType(entry['type'])
                new_val.append(self._factory_subfield(sf_type, entry))
            else:
                raise TypeError("Value entries must be of type SubBaseField or dict")

        self._sub_fields = new_val



    def add_subfield(self, sf_type: SubTableFieldType, col_name: str) -> TableSubBaseField:
        self.sub_fields.append(self._factory_subfield(sf_type, {'col_name': col_name}))
        return self.sub_fields[-1]

    def _factory_subfield(self, sf_type: SubTableFieldType,
                          data=None) -> TableSubBaseField:
        if data is None:
            data = {}
        if sf_type == SubTableFieldType.DRAG_MOLECULE:
            return TSDragMolecule(self._sr, self._layer_key, data)
        elif sf_type == SubTableFieldType.DRAG_SAMPLE:
            return TSDragSample(self._sr, self._layer_key, data)
        elif sf_type == SubTableFieldType.SELECT:
            return TSSelect(self._sr, self._layer_key, data)
        elif sf_type == SubTableFieldType.SYSTEM_DEFINED:
            return TSSystemDefined(self._sr, self._layer_key, data)
        elif sf_type == SubTableFieldType.TEXT:
            return TSText(self._sr, self._layer_key, data)


class TextFormulaField(BaseField):
    field_type = FieldType.TEXT_FORMULA

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class TextField(BaseField):
    field_type = FieldType.TEXT

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._required = False
        self._parse_json()

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError(f"required {type(value)} not supported")
        self._required = value


class TextareaField(BaseField):
    field_type = FieldType.TEXTAREA

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()


class UploadField(BaseField):
    field_type = FieldType.UPLOAD

    def __init__(self, sr: StructureRegistry, layer_key: str, json_data: dict):
        super().__init__(sr, layer_key, json_data)
        self._parse_json()



import copy
import uuid


class Workflow:

    def __init__(self, props):
        self._props = props
        self.edges = props.get('flowObject', {}).get('edges')
        self.nodes = props.get('flowObject', {}).get('nodes')
        self._wf_tree = None
        self._node_registery = {}

    @property
    def workflows(self):
        return self._wfs

    @property
    def wf_tree(self):
        if self._wf_tree is None:
            self._wf_tree = self._parse_wf()
        return self._wf_tree

    def _parse_wf_reg(self, current_node):
        start = current_node['data'] | {'id': current_node['id'], 'next': {}}
        self._node_registery[current_node['id']] = start
        for edge in self.edges:
            if edge['source'] == current_node['id'] and edge['target'] and len(edge['target']) > 1:
                for next_node in self.nodes:
                    if next_node['id'] == edge['target']:
                        start['next'][next_node['id']] = self._parse_wf_reg(next_node)
        return start

    def _parse_wf(self):
        if self.nodes is None:
            return {}
        current_node = next(x for x in self.nodes if x['id'] == '1')
        start = self._parse_wf_reg(current_node)
        return start['next']

    def _add_layer(self, properties_layers, new_layer):
        key = new_layer['key']
        i = 0
        while key in properties_layers:
            i = i + 1
            key = f"{new_layer['key']}.{i}"
        properties_layers[key] = new_layer
        return (key, new_layer)

    def _prepare_layer(self, tree_entry, wf_id=None, source_layer_id=None, wf_pos=1, pos=None):
        layer = copy.deepcopy(tree_entry['layer'])
        layer['wf_uuid'] = wf_id
        layer['ai'] = []
        layer['wf'] = wf_id is not None
        if wf_pos is not None:
            layer['wf_position'] = wf_pos
        if pos is not None:
            layer['position'] = pos
        if layer['wf']:
            layer['wf_info'] = {'node_id': tree_entry['id']}
        if source_layer_id is not None:
            layer['source_layer'] = source_layer_id
        if 'next' in tree_entry:
            max_pos = max(0, -1, *[x['position'] for x in layer['fields']])
            layer['fields'].append(
                {'type': 'wf-next', 'field': '_wf_next', 'label': 'Next', 'default': '', 'position': max_pos + 1,
                 'required': False, 'sub_fields': [], 'value': None, 'wf_options': [{'key': key,
                                                                                     'label': f"{val['layer']['label']}({val['lKey']})"}
                                                                                    for key, val in
                                                                                    tree_entry['next'].items()],
                 'text_sub_fields': []})
        return layer

    def _read_wf(self, wf_tree_id, prev_id, prop_layers):
        if wf_tree_id is not None:
            key_layer_iter = iter((key, x) for key, x in prop_layers.items() if
                                  x.get('wf_info', {}).get('node_id') == wf_tree_id and x.get('wf_info', {}).get(
                                      'source_layer') == prev_id)
            key, layer = next(key_layer_iter, (None, None))
        else:
            key = layer = None
        results = []
        if layer is not None:
            results.append(layer)
            try:
                next_wf_id = next(x['value'] for x in layer['fields'] if x['type'] == 'wf-next')
                current_node = next(x for x in self.nodes if x['id'] == next_wf_id)
                results += self._read_wf(current_node['id'], key, prop_layers)
            except:
                pass
        return results

    def init_wf(self, json_data):
        max_pos = max(-1, 0, *[elm['position'] for k, elm in json_data['properties']['layers'].items()])
        self._wfs = []
        for (id, entry) in self.wf_tree.items():
            layers = self._read_wf(id, None, json_data['properties']['layers'])
            if len(layers) == 0:
                _key, layers = self._add_layer(json_data['properties']['layers'],
                                         self._prepare_layer(entry, uuid.uuid4().__str__()))
                max_pos += 10
                layers['position'] = max_pos
                self._wfs.append({
                    'layers': [layers],
                    'pos': layers['position'],
                    'id': layers['wf_uuid']
                })
            else:
                self._wfs.append({
                    'layers': layers,
                    'pos': layers[0]['position'],
                    'id': layers[0]['wf_uuid']
                })

    def next_options(self, flat = True):
        result = []
        for wf in self._wfs:
            layer = wf['layers'][-1]
            field = next(iter(x for x in layer['fields'] if x['type'] == 'wf-next'), None)
            if field is not None:
                options = copy.deepcopy(field['wf_options'])
                for option in options:
                    option['wf_id'] = wf['id']
                    option['layer_id'] = option['key']
                    del option['key']
                result.append({
                    'label': ' -> '.join([x['label'] for x in wf['layers']]),
                    'options': options
                })
        if flat:
            flat_res = []
            for x in result:
                for o in x['options']:
                    o['source_layer'] = x['label']
                    flat_res.append(o)
            return flat_res
        return result

    def add_layer(self, json_data, layer_label, pos):
        if pos is None:
            pos = max(1,0,*[x['position'] for x in json_data['properties']['layers'].values()]) + 10
        layer = next(iter(x for x in json_data['properties_release']['layers'].values() if x['wf'] and x['label'] == layer_label), None)
        if layer is None:
            return None, None
        return self._add_layer(json_data['properties']['layers'],
                                     self._prepare_layer({'layer': layer},
                                                         wf_id=None,
                                                         pos=pos))

    def next(self, json_data, layer_id=None, wf_idx=0, wf_id=None, **kwargs) -> tuple[int,dict] | tuple[None,None]:
        if wf_id is not None:
            wf = next(x for x in self._wfs if x['id'] == wf_id)
        else:
            wf = self._wfs[wf_idx]

        layer = wf['layers'][-1]
        field = next(iter(x for x in layer['fields'] if x['type'] == 'wf-next'), None)

        if layer_id is None and field is not None and len(field['wf_options']) > 0:
            layer_id = field['wf_options'][0]['key']

        entry = self._node_registery.get(layer_id, None)
        if entry is not None:
            key, layer = self._add_layer(json_data['properties']['layers'],
                                    self._prepare_layer(entry,
                                                        wf_id='',
                                                        source_layer_id=layer['key'],
                                                        wf_pos=layer['wf_position'] + 1,
                                                        pos=layer['position']))

            if field is not None:
                field['value'] = layer_id
            wf['layers'].append(layer)
            return key, layer

        return None, None



import copy

from chemotion_api.labImotion.items.validation.registry import SchemaRegistry
from chemotion_api.labImotion.items.validation.schemas.abstract_element_schema import generic_element_schema
from chemotion_api.labImotion.items.validation.schemas.schema_properties import properties_schema as properties

generic_schema = copy.deepcopy(generic_element_schema)
generic_schema['$id'] = 'chemotion://generic/segment/draft-01'
generic_schema['title'] = "Schema for generic segment"
generic_schema['properties']['properties_template']['$ref'] = 'chemotion://generic/segment_properties/draft-01'
generic_schema['properties']['properties_release']['$ref'] = 'chemotion://generic/segment_properties/draft-01'


generic_schema['required'] += ['element_klass']

generic_schema['properties'] |= {
    "element_klass": {
        "$ref": "chemotion://generic/linked_element/draft-01"
    }
}


generic_props_schema = copy.deepcopy(properties)
generic_props_schema['$id'] = 'chemotion://generic/segment_properties/draft-01'
generic_props_schema['title'] = "Schema for generic segment properties"

SchemaRegistry.instance().register(generic_schema)
SchemaRegistry.instance().register(generic_props_schema)



from enum import Enum
from typing import Optional, Self


class UnitEnum(Enum):
    @classmethod
    def from_str(cls, str_in: str) -> Self | None:
        try:
            return cls(str_in)
        except ValueError:
            return None

    def to_default_unit(self, value: float) -> float:
        return 1

    def to_unit(self, value: float, unit: Self) -> float:
        return 1

    @classmethod
    def default_unit(cls) -> Self:
        raise NotImplementedError()


class TIME_UNITS(UnitEnum):
    SECONDS = 's'
    MINUTES = 'm'
    HOURS = 'h'

    def to_default_unit(self, value: float) -> float:
        if self == TIME_UNITS.SECONDS:
            return value / 3600
        if self == TIME_UNITS.MINUTES:
            return value / 60
        return value

    def to_unit(self, value: float, unit: Self) -> float:
        value = self.to_default_unit(value)
        if unit == TIME_UNITS.SECONDS:
            return value * 3600
        if unit == TIME_UNITS.MINUTES:
            return value * 60
        return value

    @classmethod
    def default_unit(cls) -> Self:
        return cls.HOURS

    @classmethod
    def from_str(cls, str_in: str) -> Self | None:
        if str_in == 'Second(s)':
            str_in = 's'
        return super().from_str(str_in)



class TON_UNITS(UnitEnum):
    PER_SECOND = 'TON/s'
    PER_MINUTE = 'TON/m'
    PER_HOUR = 'TON/h'

    def to_default_unit(self, value: float) -> float:
        if self == TON_UNITS.PER_SECOND:
            return value * 3600
        if self == TON_UNITS.PER_MINUTE:
            return value * 60
        return value

    def to_unit(self, value: float, unit: Self) -> float:
        value = self.to_default_unit(value)
        if unit == TON_UNITS.PER_SECOND:
            return value / 3600
        if unit == TON_UNITS.PER_MINUTE:
            return value / 60
        return value

    @classmethod
    def default_unit(cls) -> Self:
        return cls.PER_HOUR


class TEMPERATURE_UNITS(UnitEnum):
    KELVIN = 'K'
    FAHRENHEIT = '°F'
    CELSIUS = '°C'

    def to_default_unit(self, value: float) -> float:
        if self == TEMPERATURE_UNITS.FAHRENHEIT:
            return abs((value - 32) * 5 / 9 + 273.15)
        if self == TEMPERATURE_UNITS.CELSIUS:
            return abs(value + 273.15)
        return abs(value)

    def to_unit(self, value: float, unit: Self) -> float:
        value = self.to_default_unit(value)
        if unit == TEMPERATURE_UNITS.FAHRENHEIT:
            return (value - 273.15) * 9 / 5 + 32
        if unit == TEMPERATURE_UNITS.CELSIUS:
            return value - 273.15
        return value

    @classmethod
    def default_unit(cls) -> Self:
        return cls.KELVIN


class VOLUME_UNITS(UnitEnum):
    LITER = 'l'
    MILLILITER = 'ml'
    MICROLITER = 'μl'

    def to_default_unit(self, value: float) -> float:
        if self == VOLUME_UNITS.MILLILITER:
            return value / 1000
        if self == VOLUME_UNITS.MICROLITER:
            return value / (1000 ** 2)
        return value

    def to_unit(self, value: float, unit: Self) -> float:
        value = self.to_default_unit(value)
        if self == VOLUME_UNITS.MILLILITER:
            return value * 1000
        if self == VOLUME_UNITS.MICROLITER:
            return value * (1000 ** 2)
        return value

    @classmethod
    def default_unit(cls) -> Self:
        return cls.LITER


class WEIGHT_UNITS(UnitEnum):
    GRAM = 'g'
    MILLIGRAM = 'mg'
    MICROGRAM = 'μg'

    def to_default_unit(self, value: float) -> float:
        if self == WEIGHT_UNITS.MILLIGRAM:
            return value / 1000
        if self == WEIGHT_UNITS.MICROGRAM:
            return value / (1000 ** 2)
        return value

    def to_unit(self, value: float, unit: Self) -> float:
        value = self.to_default_unit(value)
        if self == WEIGHT_UNITS.MILLIGRAM:
            return value * 1000
        if self == WEIGHT_UNITS.MICROGRAM:
            return value * (1000 ** 2)
        return value

    @classmethod
    def default_unit(cls) -> Self:
        return cls.GRAM


class MOL_UNITS(UnitEnum):
    MOL = 'mol'
    MILLIMOL = 'mmol'
    MICROMOL = 'μmol'

    def to_default_unit(self, value: float) -> float:
        if self == MOL_UNITS.MILLIMOL:
            return value / 1000
        if self == MOL_UNITS.MICROMOL:
            return value / (1000 ** 2)
        return value

    def to_unit(self, value: float, unit: Self) -> float:
        value = self.to_default_unit(value)
        if self == MOL_UNITS.MILLIMOL:
            return value * 1000
        if self == MOL_UNITS.MICROMOL:
            return value * (1000 ** 2)
        return value

    @classmethod
    def default_unit(cls) -> Self:
        return cls.MOL


_unit_enums = (TIME_UNITS, TON_UNITS, TEMPERATURE_UNITS, MOL_UNITS, WEIGHT_UNITS, VOLUME_UNITS)


def convert_value_to_default(value, unit: str) -> float:
    for enum_class in _unit_enums:
        if enum_class.from_str(unit) is not None:
            default_unit = enum_class.default_unit().value
            return convert_value(value, unit, default_unit)
    raise ValueError(f'The unit {unit} is not supported!')


def convert_value(value: Optional[float], unit: str, out_unit: str) -> float | None:
    if value is None:
        return value
    if unit == out_unit:
        return value
    for enum_class in _unit_enums:
        f_unit = enum_class.from_str(unit)
        t_unit = enum_class.from_str(out_unit)
        if f_unit is not None and t_unit is not None:
            return f_unit.to_unit(value, t_unit)
    raise ValueError(f'The unit {unit} is not supported!')
